cbuffer DataBuffer
{
    uint pitch;
    uint height;
    uint color_space;
    uint transfer_function;
    float gamma;
    uint width;
};

StructuredBuffer<uint> nv12_buffer;

float3 BT601ToLinerApprox(float3 gammaColor)
{
    return gammaColor*gammaColor;
}
float3 BT601ToLiner(float3 gammaColor)
{
    float3 linearLo = gammaColor * (1.0/4.5);
    float3 linearHi = pow(abs((gammaColor + 0.099) * (1.0/1.099)), float3(1.0/0.45, 1.0/0.45, 1.0/0.45));
    float3 linearColor;
    linearColor.x = (gammaColor.x < 0.0812) ? linearLo.x : linearHi.x;
    linearColor.y = (gammaColor.y < 0.0812) ? linearLo.y : linearHi.y;
    linearColor.z = (gammaColor.z < 0.0812) ? linearLo.z : linearHi.z;
    return linearColor;
}
float3 sRGBToLinear(float3 sRGB)
{
    float3 linearLo = sRGB * (1.0/12.92);
    float3 linearHi = pow(abs((sRGB + 0.055) * (1.0/1.055)), float3(2.4, 2.4, 2.4));
    float3 linearColor;
    linearColor.r = (sRGB.r <= 0.04045) ? linearLo.r : linearHi.r;
    linearColor.g = (sRGB.g <= 0.04045) ? linearLo.g : linearHi.g;
    linearColor.b = (sRGB.b <= 0.04045) ? linearLo.b : linearHi.b;
    return linearColor;
}
float3 YCbCrToRGB(float3 YCbCr, float3 K)
{
    float3x3 M = {1.0,         1.0,                    1.0,
                  0.0,         -K.b/K.g*(2.0-2.0*K.b), 2.0-2.0*K.b,
                  2.0-2.0*K.r, -K.r/K.g*(2.0-2.0*K.r), 0.0};
    return mul(YCbCr, M);
}

float2 sampleUV(uint x, uint y)
{
    uint chroma_offset = pitch*height;

    uint uv_i = chroma_offset + y*pitch + x*2;
    uint u_j = (x&1)*2;
    uint v_j = u_j+1;

    uint uv_s = nv12_buffer[uv_i/4];

    uint u_d = (uv_s>>(u_j*8))&0xff;
    uint v_d = (uv_s>>(v_j*8))&0xff;

    float u_f = u_d;
    float v_f = v_d;

    if(color_space == 0){ // BT.601
        u_f = u_d*(224.0/255.0)+16.0;
        v_f = v_d*(224.0/255.0)+16.0;
    }
    else if(color_space == 2){ // BT.709
        u_f = u_d*(224.0/255.0)+16.0;
        v_f = v_d*(224.0/255.0)+16.0;
    }

    return float2(u_f, v_f);
}

float4 main(float4 ScreenPos : SV_Position) : SV_Target
{
    uint x = ScreenPos.x;
    uint y = ScreenPos.y;

    uint y_i = y*pitch + x;
    uint y_j = x%4;

    uint y_s = nv12_buffer[y_i/4];
    uint y_d = (y_s>>(y_j*8))&0xff;

    int offset = (y&1) ? 1 : -1;
    offset = (y==0) ? 0 : offset;
    offset = (y==(height-1)) ? 0 : offset;

    uint uv_x = x/2;
    uint uv_y = y/2;
    uint uv_xoff = min(uv_x+1, width/2-1);
    uint uv_yoff = uv_y + offset;

    float2 uv0 =
        sampleUV(uv_x, uv_y) * 0.75 +
        sampleUV(uv_x, uv_yoff) * 0.25;
    float2 uv1 =
        sampleUV(uv_xoff, uv_y) * 0.75 +
        sampleUV(uv_xoff, uv_yoff) * 0.25;

    float2 uv;
    if(x&1)
        uv = uv0 * 0.5 + uv1 * 0.5;
    else
        uv = uv0;

    float y_f = y_d;
    float u_f = uv.x;
    float v_f = uv.y;

    float3 K;
    if(color_space == 0){ // BT.601
        y_f = y_d*(219.0/255.0)+16.0;
        K = float3(0.299, 0.587, 0.114);
    }
    else if (color_space == 1){ // BT.601 FullRange
        K = float3(0.299, 0.587, 0.114);
    }
    else if(color_space == 2){ // BT.709
        y_f = y_d*(219.0/255.0)+16.0;
        K = float3(0.2126, 0.7152, 0.0722);
    }
    else if(color_space == 3){ // BT.709 FullRange
        K = float3(0.2126, 0.7152, 0.0722);
    }
    float Y = y_f*(1.0/255.0);
    float U = u_f*(1.0/255.0)-0.5; // in [-0.5, 0.5]
    float V = v_f*(1.0/255.0)-0.5; // in [-0.5, 0.5]

    float3 YUV = float3(Y, U, V);
    float3 RGB = YCbCrToRGB(YUV, K);

    if(transfer_function == 0){
        RGB = RGB;
    }
    else if(transfer_function == 1){
        RGB = BT601ToLiner(RGB);
    }
    else if(transfer_function == 2){
        RGB = sRGBToLinear(RGB);
    }
    else if(transfer_function == 3){
        RGB = pow(abs(RGB), float3(gamma, gamma, gamma));
    }

    return float4(RGB, 1.0);
}
