#include "Engine/PhongLighting.hlsl"
#include "AdaptivePlanarGrid.hlsl"

#ifdef COGS_BASIC_OCEAN_QUUX
#include "BasicOceanEffectQuux.hlsl"
#else
#include "BasicOceanEffect.hlsl"
#endif

float longitude(float2 d)
{
  float2 planeDir = normalize(-d);

  float camRelAzimuth = (0.5 / 3.14159) * acos(min(1.0, dot(planeDir, camPlaneDir)));

  if (planeDir.y * camPlaneDir.x - planeDir.x * camPlaneDir.y > 0.0) {
    camRelAzimuth = -camRelAzimuth;
  }

  return camRelAzimuth + camAzimuth;
}

float fresnel(float3 view, float3 normal)
{
  float refractionIndexRatio = 1.33;
  float R0 = pow(1.0 - refractionIndexRatio, 2.0) / pow(1.0 + refractionIndexRatio, 2.0);
  float cosAngle = 1.0 - clamp(dot(view, normal), 0, 1);

  float result = cosAngle * cosAngle;
  result = result * result;
  result = result * cosAngle;
  result = clamp((result * 1 - clamp(R0, 0, 1)) + R0, 0, 1);

  return result;
}

float fresnel2(float3 view, float3 normal)
{
  float eta = 0.66;
  float FresnelPower = 2.0;
  float F = ((1.0 - eta) * (1.0 - eta)) / ((1.0 + eta) * (1.0 + eta));

  float ratio = F + (1.0 - F) * (pow((1.0 - dot(view, normal)), FresnelPower));

  return ratio;
}

SurfaceOut surfaceFunction(VertexIn vertex)
{
  SurfaceOut Out = (SurfaceOut)0;

#ifdef COGS_BASIC_OCEAN_QUUX
  float roughness = 0.f;                //< How rough the sea should appear to be, value between 0.0 and 1.0.
  float foam = 0.0;
  float3 n = waveNormal(foam, vertex.WorldPosition.xy, vertex.wavePosition.z, roughness);
#else
  float foam = 0.0;
  float3 n = waveNormal(vertex.wavePosition.xy);
#endif
  n = n * float3(2, 2, 1);

  float3 v = normalize(inverseViewMatrix[3].xyz - vertex.WorldPosition.xyz);
  Out.Normal = normalize(n);

#if 0
  if (vertex.facing == false) {
    float3 r = refract(v, n, 1.33);

    if (r.z < 0.02) {
      // Refraction into air
      // NOTE: Disabled needless compiler warning.
#pragma warning(push)
#pragma warning(disable:4000)
      return Out;
#pragma warning(pop)
    }
  }
#endif

  float factor = -1.0;

#ifdef COGS_SHADER_GL
  factor = 1.0;
#endif

  float4 reflectedColor = float4(0, 0, 0, 1);
#if defined(COGS_BASIC_OCEAN_REFLECTION_PLANAR)
  float2 screenPosition = float2(0.5, factor * 0.5) * (vertex.screenPosition.xy / vertex.screenPosition.w) + 0.5;
  reflectedColor = PlanarReflection.Sample(PlanarReflectionSampler, screenPosition + (n.xy * 0.05)) * float4(0.7, 0.7, 0.7, 1.0);
  reflectedColor.rgb *= reflectionBrightness;
#elif defined(COGS_BASIC_OCEAN_REFLECTION_ENV_SKYBOX)
  float3 refl = reflect(-v, n);
  reflectedColor = environmentSky.SampleLevel(environmentSkySampler, refl.xzy, 0);
  reflectedColor.rgb *= reflectionBrightness * environmentBrightness;
#elif defined(COGS_BASIC_OCEAN_REFLECTION_ENV_RADIANCE)
  float3 refl = reflect(-v, n);
  reflectedColor = environmentRadiance.SampleLevel(environmentRadianceSampler, refl.xzy, 0);
  reflectedColor.rgb *= reflectionBrightness * environmentBrightness;
#endif

  if ((sceneFlags & COGS_SCENEFLAGS_SUBMERGED) == 0) {
    float fresnelTerm = fresnel(v, n);
    Out.Color = float4(waterColor.rgb * (1.0 - fresnelTerm), waterColor.a);
    Out.Emissive = (reflectedColor * fresnelTerm).rgb;
    Out.Specular = float3(0.5, 0.5, 0.5);
    Out.SpecularPower = 20;
  } else {
    float3 v_w = normalize(vertex.WorldPosition.xyz - inverseViewMatrix._m30_m31_m32);

    float fresnelTerm = fresnel(v, -n);
    float3 r = refract(v, n, 1.1);
    //Out.Color =  float4(0, 0.01, 0.02, 1);
    float w = 0;
    if (r.z < 0.0) {
      w = abs(r.z);
    }
    Out.Color.rgb = w*reflectedColor.rgb;
    Out.Color.a = 1;
  }
  //Out.Color.rgb = (float3)vertex.wavePosition.z;

  return Out;
}
