#include "Bayer.hlsl"
#include "Random.hlsl"

struct ShadowData
{
  float4 cascadeLine;
  float4 nearSplits;
  float4 farSplits;
  float4x4 lightMatrix[6];
  float shadowSampleBias;
  int numCascades;
  int enabled;
};

cbuffer ShadowBuffer
{
  ShadowData shadows[COGS_MAX_LIGHTS];
  int cascadeOffsets[COGS_MAX_LIGHTS];
};

Texture2DArray cascadedShadowMap;
SamplerComparisonState cascadedShadowSampler;

float calculateShadow(in ShadowData shadow, in float3 lightSpacePosition, in int index, in int o, in int Ns, in int Nn, in float2 eta)
{
  if (Nn == 1) {
    if ((o == 0) && (Ns==1)) {
#ifdef GLSLANG
      return cascadedShadowMap.SampleCmp(cascadedShadowSampler, float3(lightSpacePosition.xy, index), lightSpacePosition.z);
#else
      return cascadedShadowMap.SampleCmpLevelZero(cascadedShadowSampler, float3(lightSpacePosition.xy, index), lightSpacePosition.z);
#endif
    }
    return 0.f;
  }
  else {
    float rv = 0.f;

    float3 dimensions;
    cascadedShadowMap.GetDimensions(dimensions.x, dimensions.y, dimensions.z);
    float2 texelSize = 1.0 / dimensions.xy;

    if (Nn == 9) {
      float2 shifts[9] = {
        {  0.0,  0.0 },
        { -1.5,  0.0 },
        {  0.0, -1.5 },
        {  0.0,  1.5 },
        {  1.5,  0.0 },
        { -1.5, -1.5 },
        { -1.5,  1.5 },
        {  1.5, -1.5 },
        {  1.5,  1.5 }
      };

      for (int i = o; i < Ns; i++) {
        float2 tp = lightSpacePosition.xy + mul(texelSize * shifts[i], float2x2(eta.x, -eta.y,
                                                                                eta.y, eta.x));

#ifdef GLSLANG
        rv += (1.0/9.0)*cascadedShadowMap.SampleCmp(cascadedShadowSampler, float3(tp, index), lightSpacePosition.z);
#else
        rv += (1.0/9.0)*cascadedShadowMap.SampleCmpLevelZero(cascadedShadowSampler, float3(tp, index), lightSpacePosition.z);
#endif
      }
    }
    else if (Nn == 36) {

      float2 shifts[36] = {
        { -2.5, -2.5 }, { -2.5, -1.5 }, { -2.5, -0.5 }, { -2.5,  0.5 }, { -2.5,  1.5 }, { -2.5,  2.5 },
        { -1.5, -2.5 }, { -1.5, -1.5 }, { -1.5, -0.5 }, { -1.5,  0.5 }, { -1.5,  1.5 }, { -1.5,  2.5 },
        { -0.5, -2.5 }, { -0.5, -1.5 }, { -0.5, -0.5 }, { -0.5,  0.5 }, { -0.5,  1.5 }, { -0.5,  2.5 },
        {  0.5, -2.5 }, {  0.5, -1.5 }, {  0.5, -0.5 }, {  0.5,  0.5 }, {  0.5,  1.5 }, {  0.5,  2.5 },
        {  1.5, -2.5 }, {  1.5, -1.5 }, {  1.5, -0.5 }, {  1.5,  0.5 }, {  1.5,  1.5 }, {  1.5,  2.5 },
        {  2.5, -2.5 }, {  2.5, -1.5 }, {  2.5, -0.5 }, {  2.5,  0.5 }, {  2.5,  1.5 }, {  2.5,  2.5 }
      };

      for (int i = o; i < Ns; i++) {
        float2 tp = lightSpacePosition.xy + texelSize * shifts[i];

#ifdef GLSLANG
        rv += (1.0 / 36.0)*cascadedShadowMap.SampleCmp(cascadedShadowSampler, float3(tp, index), lightSpacePosition.z);
#else
        rv += (1.0 / 36.0)*cascadedShadowMap.SampleCmpLevelZero(cascadedShadowSampler, float3(tp, index), lightSpacePosition.z);
#endif
      }
    }
    return rv;
  }
}


float calculateShadowContribution(in float4 worldPosition, in float3 normal, in float3 lightDir, in int shadowIndex, in float4 screenPos)
{
#if defined(COGS_DISABLE_DIRECTIONAL_SHADOWS)
  return 1;
#endif
  ShadowData shadow = shadows[shadowIndex];

  if (!shadow.enabled) return 1;

  float viewZ = dot(shadow.cascadeLine.xyz, worldPosition.xyz) + shadow.cascadeLine.w;

  if (viewZ > shadow.farSplits[shadow.numCascades - 1]) return 1;

  int index = 0;
  for (int c = shadow.numCascades - 1; c >= 0; --c) {
    if (viewZ < shadow.farSplits[c]) {
      index = c;
    }
  }

  int lightIndex = index + cascadeOffsets[shadowIndex];

  float4 lightSpacePosition = mul(float4(worldPosition.xyz, 1), shadow.lightMatrix[index]);
  float3 pos_l = (1.0 / lightSpacePosition.w)*lightSpacePosition.xyz;
  pos_l.z = saturate(pos_l.z - shadow.shadowSampleBias);

#if defined(COGS_SOFT_SHADOWS_HIGH)
  int Nn = 36;
#elif defined(COGS_SOFT_SHADOWS_LOW)
  int Nn = 9;
#else
  int Nn = 1;
#endif

  float2 eta;
#if COGS_SOFT_SHADOWS_JITTER==1
  sincos((6.28318530718 / 4294967295.0)*wangHash(1664525 * ((uint)screenPos.y) + (uint)screenPos.x), eta.y, eta.x);
#else
  eta = float2(1, 0);
#endif

  float fade = viewZ - shadow.nearSplits[index];
  if (0.f < fade && index < shadow.numCascades - 1) {
    fade = (1.f / (shadow.farSplits[index] - shadow.nearSplits[index])) * fade;

    int Ns = min(Nn, Nn * fade + (1.0 / 15.0)*bayer4((uint2)screenPos.xy));

    float acc = calculateShadow(shadow, pos_l, lightIndex, Ns, Nn, Nn, eta);
    if (index < shadow.numCascades - 1) {
      float4 lightSpacePosition = mul(float4(worldPosition.xyz, 1), shadow.lightMatrix[index + 1]);
      float3 pos_l = (1.0 / lightSpacePosition.w)*lightSpacePosition.xyz;
      pos_l.z = saturate(pos_l.z - shadow.shadowSampleBias);
      acc += calculateShadow(shadow, pos_l, lightIndex + 1, 0, Ns, Nn, eta);
    }
    else {
      acc += float(Ns) / Nn;
    }
    return acc;
  }
  else {
    return calculateShadow(shadow, pos_l, lightIndex, 0, Nn, Nn, eta);
  }
}

float calculateShadowContribution(in float viewZ, in float4 worldPositionA, in float4 worldPositionB, in float3 lightDir, in int shadowIndex, in float4 screenPos)
{
#if defined(COGS_DISABLE_DIRECTIONAL_SHADOWS)
  return 1;
#endif
  ShadowData shadow = shadows[shadowIndex];

  if (!shadow.enabled) return 1;

  if (viewZ > shadow.farSplits[shadow.numCascades - 1]) return 1;

  int index = 0;
  for (int c = shadow.numCascades - 1; c >= 0; --c) {
    if (viewZ < shadow.farSplits[c]) {
      index = c;
    }
  }

  float4 lightSpacePositionA = mul(worldPositionA, shadow.lightMatrix[index]);
  lightSpacePositionA.xyz = (1.f / lightSpacePositionA.w)*lightSpacePositionA.xyz;
  lightSpacePositionA.z -= shadow.shadowSampleBias;

  float4 lightSpacePositionB = mul(worldPositionB, shadow.lightMatrix[index]);
  lightSpacePositionB.xyz = (1.f / lightSpacePositionB.w)*lightSpacePositionB.xyz;
  lightSpacePositionB.z -= shadow.shadowSampleBias;

  float3 dimensions;
  cascadedShadowMap.GetDimensions(dimensions.x, dimensions.y, dimensions.z);
  float2 texelSize = 1.0 / dimensions.xy;

  int N = max(1, min(25, length(dimensions.xy*(lightSpacePositionA.xy - lightSpacePositionB.xy))));
  float accum = 0;
  float weight = 0.0;
  for (int i = 0; i < N; i++) {
    float3 lightSpacePosition = lerp(lightSpacePositionA.xyz, lightSpacePositionB.xyz, (i + 0.5) / N);
    lightSpacePosition.z = saturate(lightSpacePosition.z);
    //accum += cascadedShadowMap.SampleCmpLevelZero(cascadedShadowSampler, float3(lightSpacePosition.xy, index), lightSpacePosition.z);
    accum += cascadedShadowMap.SampleCmp(cascadedShadowSampler, float3(lightSpacePosition.xy, index), lightSpacePosition.z);
    weight += 1.0;
  }

  return accum / weight;
}
