Texture2D colorTexture;
Texture2D depthTexture;
Texture2D normalTexture;

SamplerState linearSampler;

struct VertexIn
{
  float4 Position : SV_Position;
  float4 ScreenPosition : TEXCOORD0;
  float2 TexCoords : TEXCOORD1;
};

struct PsOutput
{
  float4 Color : SV_Target;
  float Depth : SV_Depth;
};

cbuffer BasicOutlineParameters
{
  float outlineBrightness;
  float outlineStrength;
  float normalThreshold;
  float depthThreshold;
  float outlineFadeThreshold;
};

float normalOutline(in float2 texCoords, in float2 offset, in float2 pixelSize)
{
  float3 normal00 = normalTexture.Sample(linearSampler, offset + texCoords + pixelSize * float2(0, 0)).xyz;
  float3 normal10 = normalTexture.Sample(linearSampler, offset + texCoords + pixelSize * float2(1, 0)).xyz;
  float3 normal01 = normalTexture.Sample(linearSampler, offset + texCoords + pixelSize * float2(0, 1)).xyz;

  bool normalOffset = dot(normal00, normal10) < normalThreshold || dot(normal00, normal01) < normalThreshold;

  return normalOffset ? 1 : 0;
}

float depthOutline(in float depth, in float2 texCoords, in float2 offset, in float2 pixelSize)
{
  float depthL = depthTexture.Sample(linearSampler, offset + texCoords + pixelSize * float2(-1, 0)).x;
  float depthR = depthTexture.Sample(linearSampler, offset + texCoords + pixelSize * float2(1, 0)).x;
  float depthB = depthTexture.Sample(linearSampler, offset + texCoords + pixelSize * float2(0, -1)).x;
  float depthT = depthTexture.Sample(linearSampler, offset + texCoords + pixelSize * float2(0, 1)).x;

  float diffL = depth - depthL;
  float diffR = depthR - depth;
  float diffB = depth - depthB;
  float diffT = depthT - depth;

  return ((diffR - diffL) > depthThreshold || (diffT - diffB) > depthThreshold) ? 1 : 0;
}

PsOutput main(VertexIn In)
{
  const float4 color = colorTexture.Load(int3(In.Position.xy, 0));
  const float depth = depthTexture.Load(int3(In.Position.xy, 0)).x;

  float outline = 0.0;
  float o = 0.25;

  float2 dimensions;
  normalTexture.GetDimensions(dimensions.x, dimensions.y);
  float2 pixelSize = 1.0 / dimensions;

  outline += normalOutline(In.TexCoords, float2(0, 0) * pixelSize, pixelSize);
  outline += normalOutline(In.TexCoords, float2(o, 0) * pixelSize, pixelSize);
  outline += normalOutline(In.TexCoords, float2(o, o) * pixelSize, pixelSize);
  outline += normalOutline(In.TexCoords, float2(0, o) * pixelSize, pixelSize);

  depthTexture.GetDimensions(dimensions.x, dimensions.y);
  pixelSize = 1.0 / dimensions;

  outline += depthOutline(depth, In.TexCoords, float2(0, 0) * pixelSize, pixelSize);
  outline += depthOutline(depth, In.TexCoords, float2(o, 0) * pixelSize, pixelSize);
  outline += depthOutline(depth, In.TexCoords, float2(o, o) * pixelSize, pixelSize);
  outline += depthOutline(depth, In.TexCoords, float2(0, o) * pixelSize, pixelSize);

  outline /= 8.0;

  outline = saturate(lerp(outline, 0, smoothstep(outlineFadeThreshold, 1, depth))) * outlineStrength;

  PsOutput Out = (PsOutput)0;
  Out.Color = lerp(color, float4(outlineBrightness, outlineBrightness, outlineBrightness, 1), outline);
  Out.Depth = depth;

  return Out;
}
